from sqlalchemy.orm import sessionmaker
from database.engine import get_engine
from sqlalchemy.sql import select


class Repository:
    def __init__(self, model_type, db_name="embeddings"):
        self.engine = get_engine(db_name=db_name)
        Session = sessionmaker(bind=self.engine)
        self.session = Session
        self.model_type = model_type

    def create(self, **kwargs):
        entity = self.model_type(**kwargs)
        with self.session() as session:
            session.add(entity)
            session.commit()
        return entity

    def get_by_id(self, id):
        with self.session(self.engine) as session:
            entity = session.query(self.model_type).get(id)
        return entity

    def get(self):
        with self.session() as session:
            entity = session.query(self.model_type).first()
        return entity

    def search(self, embedding):
        with self.session() as session:
            entities = session.scalars(
                select(self.model_type)
                .order_by(self.model_type.embedding.l2_distance(embedding))
                .limit(5)
            ).all()
        return entities

    def update(self, id, **kwargs):
        with self.session() as session:
            entity = session.query(self.model_type).get(id)
            for key, value in kwargs.items():
                setattr(entity, key, value)
            session.commit()
        return entity

    def delete(self, id):
        with self.session() as session:
            entity = session.query(self.model_type).get(id)
            session.delete(entity)
            session.commit()
        return entity

    def delete_all(self):
        with self.session() as session:
            session.query(self.model_type).delete()
            session.commit()
        return True



def get_system_prompt(entity):
    return f"""
    You are a helpful customer support representative for {entity}
    that answers questions based on given context and question, you will also be given a url where the information came from.
    Under no circumstances should you give an answer that is not in the context.
    It's extremely important that you are polite and helpful.
    If they ask you if you are a bot, you should say that you are an ai assistant but dont mention who trained you.
    Ignore any thing to do with Sidus Heroes (SIDUS) and just give your name the data is not relevant and needs to be removed.
    Do not under any circumstances refer to "The article" or "Context" use this for your answer but don't mention it.
  """.strip()


def get_system_prompt_test(entity):
    return f"""
      You are a helpful customer support representative for {entity}
      that answers questions based on given context and question, you will also be given a url where the information came from.
    """


def get_context_prompt(url, context, question):
    return f"""
      Url: {url}
      Context: {context}
      Question: {question}
    """.strip()



from sqlalchemy import Column, Integer, String
from pgvector.sqlalchemy import Vector
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import mapped_column

from . import Base


class Document(Base):
    __tablename__ = "documents"
    id = Column(Integer, primary_key=True)
    doc_id = Column(String)
    doc_text = Column(String)
    doc_url = Column(String)
    doc_title = Column(String)
    embedding = mapped_column(Vector(768))



import os
import openai
import time
import threading
import sys
import itertools


OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
openai.api_key = OPENAI_API_KEY

model_engine = "gpt-3.5-turbo"


def send_system_message(messages):
    response = openai.ChatCompletion.create(model=model_engine, messages=messages)
    return response


def chat(prompts):
    messages = [{"role": "system", "content": prompts.system_prompt("The company")}]
    send_system_message(messages)
    while True:
        reply = None
        user_input = input("👤: ")
        message = prompts.context_prompt(user_input)

        if message == "quit":
            break

        if message == "clear":
            messages = []
            send_system_message(messages)
            continue

        if message == "history":
            print(messages)
            continue

        if message == "help":
            print(
                "Type 'quit' to exit the chat, 'clear' to clear the chat history, 'history' to view the chat history, and 'help' to view this message."
            )
            continue

        if message:
            messages.append({"role": "user", "content": message})
            done = False

            def animate():
                for c in itertools.cycle(["|", "/", "-", "\\"]):
                    if done:
                        break
                    sys.stdout.write("\rloading " + c)
                    sys.stdout.flush()
                    sys.stdout.write("\r")
                    time.sleep(0.1)

            t = threading.Thread(target=animate)
            t.start()

            completion = openai.ChatCompletion.create(
                model=model_engine,
                messages=messages,
            )
            reply = completion["choices"][0]["message"]["content"]
            print(f"🤖: {reply}")
            messages.append({"role": "assistant", "content": reply})

            done = True


if __name__ == "__main__":
    chat()



import json
import os
import unittest

import sys

sys.path.append("src")

from src.quintus import Quintus

model_name = os.environ.get("EMBEDDING_MODEL")


class TestStore(unittest.TestCase):
    def setUp(self):
        self.data_dir = "test_data"
        self.parser = "html.parser"

        self.document_store = Quintus(
            data_dir=self.data_dir,
            db_name="test_embeddings",
        )

        if not os.path.exists(self.data_dir):
            os.mkdir(self.data_dir)

        samples = [
            {
                "id": 123,
                "body": "<html><body>This is a sample article.</body></html>",
            },
            {
                "id": 456,
                "body": "<html><body>This is another document.</body></html>",
            },
        ]

        for sample in samples:
            with open(os.path.join(self.data_dir, f'{sample["id"]}.json'), "w") as f:
                json.dump(sample, f)

    def tearDown(self):
        self.document_store.document_repository.delete_all()

    def test_index_documents(self):
        self.document_store.index()
        docs = self.document_store.search("article")
        self.assertEqual(len(docs), 2)


if __name__ == "__main__":
    unittest.main()



import re
from quintus import Quintus

quintus = Quintus()

url = "https://postman.zendesk.com/api/v2/help_center/en-us/articles.json"


def name_filter(article):
    return not not re.search(r"(.*?)\s*\\u2014\s*Delisted", article["title"])


filters = [name_filter]

quintus.scrape(url, filters=filters)



from utils.processor import Processor
from templates.templates import (
    get_system_prompt,
    get_system_prompt_test,
    get_context_prompt,
)


class Prompts:
    def __init__(self, quintus):
        self.quintus = quintus

    def process(self, question):
        max_prompt_length = 1024
        results = self.quintus.search(question)
        text = results[0].doc_text
        url = results[0].doc_url
        processor = Processor()
        context = processor.html_to_text(text)
        context = context[:max_prompt_length]
        return url, context

    def context_prompt(self, question):
        url, context = self.process(question)
        return get_context_prompt(question, context, url)

    def system_prompt(self, entity):
        return get_system_prompt(entity)

    def system_prompt_test(self, entity):
        return get_system_prompt_test(entity)



import torch
import copy
from torch.nn import functional as F
import pydash as _
import torch
import asyncio


async def generate(input_text, tokenizer, model, device, **kwargs):
    input_ids = tokenizer.encode(input_text, return_tensors="pt").to(device)
    max_new_tokens = kwargs.pop("max_new_tokens", 2048)
    batch_size = input_ids.shape[0]
    input_length = input_ids.shape[-1]
    config = model.config
    config = copy.deepcopy(config)
    pad_token_id = config.pad_token_id
    eos_token_id = config.eos_token_id
    consecutive_eos_tokens = 0
    target_consecutive_eos_tokens = 5
    consecutive_line_breaks = 0
    target_consecutive_line_breaks = 2
    line_break_token = 13

    kwargs["output_attentions"] = False
    kwargs["output_hidden_states"] = False
    kwargs["use_cache"] = config.use_cache

    if isinstance(eos_token_id, int):
        eos_token_id = [eos_token_id]

    incomplete_sequences = input_ids.new_ones(batch_size)

    while True:
        outputs = model(input_ids, **kwargs, return_dict=True)

        logits = outputs.logits[:, -1, :]

        with torch.inference_mode():
            logits = logits

        probs = F.softmax(logits.float(), dim=-1)

        tokens = torch.argmax(probs, dim=-1)[:, None]

        token_logprobs = torch.gather(probs, 1, tokens)

        token_logprobs = torch.log(token_logprobs + 1e-7).squeeze(1)

        tokens = tokens.squeeze(1)

        if pad_token_id is not None:
            tokens = tokens * incomplete_sequences + pad_token_id * (
                1 - incomplete_sequences
            )

        input_ids = torch.cat([input_ids, tokens[:, None]], dim=-1)

        if eos_token_id is not None:
            not_eos = torch.all(
                torch.tensor([tokens != i for i in eos_token_id]), dim=0
            ).to(device)
            incomplete_sequences = incomplete_sequences.mul(not_eos.long())

            if tokens.item() == eos_token_id:
                consecutive_eos_tokens += 1
            else:
                consecutive_eos_tokens = 0

            if tokens.item() == line_break_token:
                consecutive_line_breaks += 1
            else:
                consecutive_line_breaks = 0

            if (
                consecutive_eos_tokens >= target_consecutive_eos_tokens
                or consecutive_line_breaks >= target_consecutive_line_breaks
            ):
                break

        status = incomplete_sequences.clone()

        if input_ids.shape[-1] - input_length >= max_new_tokens:
            status = 0 - status

        is_finished = incomplete_sequences.max() == 0

        await asyncio.sleep(0.01)

        yield tokens, is_finished

        if status.max() <= 0:
            break



import os
import json
import requests
import pydash as _


class Scraper:
    def __init__(self, data_dir):
        self.url = None
        self.filters = None
        self.data_dir = data_dir
        if not os.path.exists(data_dir):
            os.makedirs(data_dir)

    def filter_result(self, article) -> bool:
        return True in [f(article) for f in self.filters]

    def write_article(self, article):
        article_id = article["id"]
        with open(
            os.path.join(self.data_dir, f"article{article_id}.json"), "w"
        ) as article_file:
            article_file.write(
                json.dumps(
                    {
                        "id": _.get(article, "id"),
                        "title": _.get(article, "title"),
                        "body": _.get(article, "body"),
                        "url": _.get(article, "url"),
                    }
                )
            )

    def get_page(self, page_number):
        page_url = f"{self.url}?page={page_number}"
        response = requests.get(page_url)
        data = response.json()
        return data

    def scrape(self, url, filters):
        self.url = url
        self.filters = filters
        page_number = 1
        while True:
            data = self.get_page(page_number)
            articles = data["articles"]
            if len(articles) == 0:
                break
            for article in articles:
                if not self.filter_result(article):
                    self.write_article(article)
            page_number += 1



from sqlalchemy import Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

from . import Base


class LocalModelConfig(Base):
    __tablename__ = "local_model_config"
    id = Column(Integer, primary_key=True)
    model_name_or_path = Column(String)
    ft_model_name_or_path = Column(String)
    temperature = Column(Integer)
    top_p = Column(Integer)
    top_k = Column(Integer)
    num_beams = Column(Integer)
    max_new_tokens = Column(Integer)
    openai_api_key = Column(String)



from sqlalchemy import Column, Integer, String, Boolean, DateTime
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class PostmanArticle(Base):
    __tablename__ = "postman_articles"

    id = Column(Integer, primary_key=True)
    url = Column(String)
    html_url = Column(String)
    author_id = Column(Integer)
    created_at = Column(DateTime)
    updated_at = Column(DateTime)
    name = Column(String)
    title = Column(String)
    locale = Column(String)
    edited_at = Column(DateTime)
    body = Column(String)



from schema.config import LocalModelConfig
from database.repository import Repository


def seed_config():
    repository = Repository(LocalModelConfig)

    repository.create(
        model_name_or_path="models/llama-7b",
        ft_model_name_or_path="models/llama-7b-ft",
        temperature=0.2,
        top_p=0.9,
        top_k=50,
        num_beams=4,
        max_new_tokens=512,
    )


def seed():
    seed_config()


if __name__ == "__main__":
    seed()



import torch
import asyncio
import pydash as _

from inference.local.generate import generate


async def stream(prompt, tokenizer, model, device):
    sequence = torch.empty(0).to(device)
    last_printed_index = 0

    async for token, is_finished in generate(prompt, tokenizer, model, device):
        sequence = torch.cat((sequence, token))
        if token < tokenizer.vocab_size:
            text = tokenizer.decode(sequence, skip_special_tokens=True)
            new_text = text[last_printed_index:]
            yield new_text
            last_printed_index = len(text)
            await asyncio.sleep(0.01)

        if is_finished:
            break



from sentence_transformers import SentenceTransformer
import torch


class Encoder:
    def __init__(self, model_name):
        self.model = SentenceTransformer(model_name)
        self.device = "cuda:0" if torch.cuda.is_available() else "cpu"
        self.model.to(self.device)

    def encode(self, document):
        with torch.no_grad():
            embeddings = self.model.encode(document, device=self.device)
            return embeddings



from bs4 import BeautifulSoup
from pydash import join
import pydash as _
import spacy


class Processor:
    def __init__(self):
        spacy.prefer_gpu()
        self.nlp = spacy.load("en_core_web_sm")
        self.parser = "html.parser"

    def html_to_text(self, html):
        soup = BeautifulSoup(html, self.parser)
        doc_text = soup.get_text()
        return doc_text

    def process(self, text):
        doc_text = join(
            [
                token.text
                for token in self.nlp(text)
                if not token.is_stop and not token.is_punct
            ],
            " ",
        )
        return doc_text



import torch
from transformers import LlamaTokenizer, LlamaForCausalLM
from peft import PeftModel
from pathlib import Path


def get_model(config):
    model_path = Path(__file__).parent / config.model_name_or_path
    ft_model_path = Path(__file__).parent / config.ft_model_name_or_path
    tokenizer = LlamaTokenizer.from_pretrained(model_path)

    model = LlamaForCausalLM.from_pretrained(
        model_path,
        load_in_8bit=True,
        torch_dtype=torch.float16,
        device_map="auto",
    )

    model = PeftModel.from_pretrained(
        model, ft_model_path, device_map={"": 0}, torch_dtype=torch.float16
    )

    return model, tokenizer



from sqlalchemy import Column, Integer, String
from . import Base


class Provider(Base):
    __tablename__ = "providers"
    id = Column(Integer, primary_key=True)
    name = Column(String)


class OpenAIProvider(Provider):
    __tablename__ = "openai_providers"
    model_engine = Column(String)



from setuptools import setup

setup(
    name="quintus",
    version="0.1",
    description="Document Embedding and Inference",
    author="rjmacarthy",
    author_email="rjmacarthy@protonmail.com",
    url="https://github.com/rjmacarthy/quintus",
    package_dir={"": "lib"},
    install_requires=[
        "beautifulsoup4",
        "pgvector",
        "json",
        "numpy",
        "pathlib",
        "pydash",
        "sentence-transformers",
        "spacy",
        "torch",
        "tqdm",
        "unittest",
        "openai",
    ],
)



from pathlib import Path
from tqdm import tqdm
import json
import pydash as _
import torch
import os

from database.repository import Repository
from database.schema.document import Document
from utils.encoder import Encoder
from utils.processor import Processor
from utils.scraper import Scraper
from inference.providers.openai import chat as openai_chat
from inference.providers.provider_names import ProviderNames
from templates.prompts import Prompts


class Quintus:
    def __init__(
        self,
        *,
        model_name=os.environ.get("EMBEDDING_MODEL")
        or "sentence-transformers/all-mpnet-base-v2",
        data_dir="data",
        parser="html.parser",
        db_name="embeddings",
    ):
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.parser = parser
        self.data_dir = Path(data_dir)
        self.encoder = Encoder(model_name)
        self.processor = Processor()
        self.document_repository = Repository(Document, db_name)
        self.scraper = Scraper(self.data_dir)
        self.prompts = Prompts(self)
        self.model_name = model_name
        self.providers = ProviderNames

    def get_provider(self, provider):
        if provider == ProviderNames.OPEN_AI.value:
            return openai_chat
        else:
            raise Exception(f"Provider {provider} not found")

    def scrape(self, url, filters):
        print(f"Scraping {url} with filters {filters}...")
        self.scraper.scrape(url, filters)
        print("Done scraping.")

    def search(self, query):
        with torch.no_grad():
            return self.document_repository.search(self.encoder.encode(query))

    def chat(self, provider):
        return self.get_provider(provider)(self.prompts)

    def injest(self):
        json_files = sorted(list(self.data_dir.glob("*.json")))
        for _, file in enumerate(tqdm(json_files, desc="Processing JSON files")):
            with open(file) as f:
                doc = json.load(f)
                doc_text = self.processor.html_to_text(doc["body"])
                embedding = self.encoder.encode(doc_text)
            self.document_repository.create(
                doc_id=doc["id"],
                doc_text=doc["body"],
                doc_title=doc["title"],
                doc_url=doc["url"],
                embedding=embedding,
            )



import os
from sqlalchemy import create_engine, Engine
from sqlalchemy_utils import create_database, database_exists
from sqlalchemy.sql import text
from sqlalchemy.ext.declarative import declarative_base

from database.schema import Base

PGPASSWORD = os.environ.get("PGPASSWORD") or "password"
PGUSER = os.environ.get("PGUSER") or "postgres"
PGDATABASE = os.environ.get("PGDATABASE") or "embeddings"
PGHOST = os.environ.get("PGHOST") or "localhost"
PGPORT = os.environ.get("PGPORT") or 5432


def get_uri(db_name):
    return f"postgresql://{PGUSER}:{PGPASSWORD}@{PGHOST}:{PGPORT}/{db_name}"


def get_engine(
    db_name: str = PGDATABASE,
):
    engine: Engine = create_engine(get_uri(db_name))

    if not database_exists(engine.url):
        create_database(engine.url)

    with engine.connect() as con:
        con.execute(text("CREATE EXTENSION IF NOT EXISTS vector"))
        con.commit()

    Base.metadata.create_all(engine)

    return engine
